import math
import sys
import traceback
from pathlib import Path
from typing import List, Optional

from sohstationviewer.controller.util import display_tracking_info
from sohstationviewer.view.plotting.gps_plot.gps_point import GPSPoint
from sohstationviewer.model.mseed.mseed import MSeed

from PySide2 import QtWidgets, QtCore
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as Canvas
from matplotlib.figure import Figure
from matplotlib.axes import Axes

from sohstationviewer.view.util.enums import LogType


class GPSWidget(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget,
                 tracking_box: QtWidgets.QTextBrowser):
        super().__init__(parent)
        self.tracking_box = tracking_box
        self.gps_points: Optional[List[GPSPoint]] = None

        self.fig = Figure(figsize=(6, 6), dpi=100)
        self.canvas = Canvas(self.fig)
        self.canvas.mpl_connect('pick_event', self.on_pick_event)

        self.ax: Axes = self.fig.add_axes((0, 0, 1, 1))
        self.ax.set_facecolor("dimgray")
        # Force the plot to be square.
        self.ax.set_box_aspect(1)
        # Remove all traces of the axes, leaving only the plot.
        self.ax.set_yticks([])
        self.ax.set_xticks([])
        self.ax.spines['right'].set_visible(False)
        self.ax.spines['left'].set_visible(False)
        self.ax.spines['top'].set_visible(False)
        self.ax.spines['bottom'].set_visible(False)

    @QtCore.Slot()
    def plot_gps(self):
        """
        Plot the GPS points onto the canvas. Display the latitude and longitude
        range in meters and adjust the x and y limits of the plot so that the
        axis is square.
        """
        all_latitudes = [point.latitude for point in self.gps_points]
        all_longitudes = [point.longitude for point in self.gps_points]
        max_latitude = max(all_latitudes)
        min_latitude = min(all_latitudes)
        max_longitude = max(all_longitudes)
        min_longitude = min(all_longitudes)

        latitude_range = convert_latitude_degree_to_meter(
            max_latitude - min_latitude
        )
        # qpeek uses the average latitude in this conversion, so we copy it.
        longitude_range = convert_longitude_degree_to_meter(
            max_longitude - min_longitude,
            (max_latitude + min_latitude) / 2
        )

        self.ax.plot(all_longitudes, all_latitudes, 'ws', markersize=4,
                     markeredgecolor='black', picker=True, pickradius=4)
        range_text = (f'Lat range: {latitude_range:.2f}m    '
                      f'Long range: {longitude_range:.2f}m')
        self.ax.text(0.01, 0.96, range_text, color='white', size=10,
                     transform=self.ax.transAxes
                     )

        # Stretch the shorter axis so that it is as long as the longer axis.
        latitude_range = max_latitude - min_latitude
        longitude_range = max_longitude - min_longitude
        if latitude_range > longitude_range:
            longitude_average = (min_longitude + max_longitude) / 2
            min_longitude = longitude_average - latitude_range / 2
            max_longitude = longitude_average + latitude_range / 2
            self.ax.set_xlim(min_longitude, max_longitude)
        else:
            latitude_average = (min_latitude + max_latitude) / 2
            min_latitude = latitude_average - longitude_range / 2
            max_latitude = latitude_average + longitude_range / 2
            self.ax.set_ylim(min_latitude, max_latitude)

        self.canvas.draw()
        self.repaint()
        msg = f'Plotted {len(self.gps_points)} points.'
        display_tracking_info(self.tracking_box, msg)

    def on_pick_event(self, event):
        """
        On a GPS point being picked, display the data of that point onto the
        tracking box.

        :param event: the pick event
        """
        # index of the clicked point on the plot
        click_plot_index = event.ind[0]
        picked_point = self.gps_points[click_plot_index]

        lat_dir = 'N' if picked_point.latitude > 0 else 'S'
        long_dir = 'E' if picked_point.longitude > 0 else 'W'
        meta_separator = '&nbsp;' * 22
        loc_separator = '&nbsp;' * 10
        msg = (
            f'Mark: {picked_point.last_timemark}{meta_separator}'
            f'Fix: {picked_point.fix_type}{meta_separator}'
            f'Sats: {picked_point.num_satellite_used}<br>'
            f'Lat: {lat_dir}{abs(picked_point.latitude):.6f}{loc_separator}'
            f'Long: {long_dir}{abs(picked_point.longitude):.6f}{loc_separator}'
            f'Elev: {picked_point.height}{picked_point.height_unit}'
        )
        display_tracking_info(self.tracking_box, msg)


class GPSDialog(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__()
        self.parent = parent

        self.data_path: Path = None

        self.info_text_browser = QtWidgets.QTextBrowser(self)
        self.plotting_widget = GPSWidget(self, self.info_text_browser)

        self.read_button = QtWidgets.QPushButton('Read/Plot', self)
        self.export_button = QtWidgets.QPushButton('Export', self)
        self.close_button = QtWidgets.QPushButton('Close', self)

        self.setup_ui()

    def setup_ui(self):
        """
        Set up the user interface of the dialog.
        """
        # The height of the dialog is chosen so that the GPS plot is flush with
        # the edges of the dialog.
        dialog_height = 518
        self.setGeometry(300, 300, 400, dialog_height)
        self.setWindowTitle("GPS Plot")

        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)

        main_layout.addWidget(self.plotting_widget.canvas)

        # Add a 1-pixel high widget to the main layout so that there is a
        # visible border between the GPS plot and the button. QTextBrowser
        # is used because it is known to have a visible border.
        invisible_widget = QtWidgets.QTextBrowser()
        invisible_widget.setFixedHeight(1)
        main_layout.addWidget(invisible_widget)

        button_layout = QtWidgets.QHBoxLayout()
        button_layout.setContentsMargins(100, 0, 100, 0)
        button_layout.setStretch(1, 3)
        button_layout.setSpacing(10)

        self.read_button.clicked.connect(self.plotting_widget.plot_gps)
        button_layout.addWidget(self.read_button)

        self.export_button.clicked.connect(self.export_gps_points)
        button_layout.addWidget(self.export_button)

        self.close_button.setStyleSheet('QPushButton {color: red;}')
        self.close_button.clicked.connect(self.close)
        button_layout.addWidget(self.close_button)

        bottom_layout = QtWidgets.QVBoxLayout()
        bottom_layout.addLayout(button_layout)
        self.info_text_browser.setFixedHeight(42)
        self.info_text_browser.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)
        self.info_text_browser.setVerticalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)
        self.info_text_browser.verticalScrollBar().setDisabled(True)
        self.info_text_browser.horizontalScrollBar().setDisabled(True)
        bottom_layout.addWidget(self.info_text_browser)
        main_layout.addLayout(bottom_layout)

    @property
    def gps_points(self):
        return self.plotting_widget.gps_points

    @gps_points.setter
    def gps_points(self, points):
        self.plotting_widget.gps_points = [point
                                           for point in points
                                           if not point.is_bad_point()]

    def export_gps_points(self):
        """
        Export the data of the GPS points to a file. The file will be in the
        same directory as the data folder.

        The first line in the export file contains information about the data
        set. Each subsequent line has the form
        <YYYY-MM-DD HH:MM:SS> <fix type> <no. sats. used> <latitude:.6f> <longitude:.6f> <height:.6f>  # noqa
        and is tab-separated.
        """
        self.info_text_browser.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarAsNeeded)
        self.info_text_browser.setVerticalScrollBarPolicy(
            QtCore.Qt.ScrollBarAsNeeded)
        self.info_text_browser.verticalScrollBar().setDisabled(False)
        self.info_text_browser.horizontalScrollBar().setDisabled(False)

        folder_name = self.data_path.name
        outside_folder = self.data_path.parent
        export_file_path = outside_folder / f'{folder_name}.gps.dat'

        try:
            with open(export_file_path, 'w+') as outfile:
                outfile.write(f'# GPS data points for {folder_name}\n')
                for point in self.gps_points:
                    if point.is_bad_point():
                        continue
                    line = (
                        f'{point.last_timemark}\t{point.fix_type}\t'
                        f'{point.num_satellite_used}\t{point.latitude:+.6f}\t'
                        f'{point.longitude:+.6f}\t{point.height:.6f}\n'
                    )
                    outfile.write(line)
        except OSError:
            err_msg = traceback.format_exc()
            msg = f"Can't export to file due to error: {err_msg}"
            display_tracking_info(self.info_text_browser, msg,
                                  LogType.ERROR)
        else:
            msg = f'Successfully exported to {export_file_path}'
            display_tracking_info(self.info_text_browser, msg, LogType.INFO)


# We use the WGS-84's version of the Earth ellipsoid as a reference point for
# converting latitude and longitude differences from degree to meter.
# The constants below are obtained from here
# https://en.wikipedia.org/wiki/Earth_ellipsoid#Historical_Earth_ellipsoids.
# In order to use other models of the Earth ellipsoid, simply change these
# constants.
POLAR_CIRCUMFERENCE = 40007863
EQUATORIAL_CIRCUMFERENCE = 40075017


def convert_latitude_degree_to_meter(lat: float) -> float:
    """
    Convert the given latitude from degree to meter.

    :param lat: latitude given in degree
    :return: the given latitude converted to meter
    """
    # A whole circumference is 360 degrees, so we can get the length of one
    # degree by dividing the circumference by 360.
    lat_degree_length_in_meter = POLAR_CIRCUMFERENCE / 360
    return lat * lat_degree_length_in_meter


def convert_longitude_degree_to_meter(long: float, lat: float) -> float:
    """
    Convert the given longitude from degree to meter. Need to adjust for
    latitude because the length of a longitude degree changes with the
    latitude.

    :param long: longitude given in degree
    :param lat: the latitude to adjust for
    :return: the given longitude converted to meter
    """
    # A whole circumference is 360 degrees, so we can get the length of one
    # degree by dividing the circumference by 360.
    long_degree_length_in_meter = EQUATORIAL_CIRCUMFERENCE / 360
    adjustment_for_latitude = math.cos(math.radians(lat))
    return long * long_degree_length_in_meter * adjustment_for_latitude


if __name__ == '__main__':
    import os

    os.chdir('/')
    app = QtWidgets.QApplication(sys.argv)

    data_path = '/test_data/Q330-sample'  # noqa: E501
    # data_path = '/Users/kle/Documents/SOHView data/Q330/Q330_5281.sdr'
    data = MSeed(QtWidgets.QTextBrowser(), data_path)

    wnd = GPSDialog()
    wnd.gps_points = data.gps_points
    wnd.data_path = Path(data_path)
    wnd.show()

    sys.exit(app.exec_())
